<?php /*     ___ ___ _______ _______ ________ __ __     __ 
            |   |   |   _   |_     _|  |  |  |__|  |--.|__|
             \     /|       | |   | |  |  |  |  |    < |  |
              |___| |___|___| |___| |________|__|__|__||__|

Copyright (c) 2011-2017, mappy. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that redistributions of source code retain
    the above copyright notice and this license text.

URL rewriting for nginx;
location /wiki {
	try_files $uri $uri/ /wiki/index.php?q=$uri&$args;
}

Supports PHP 5.3+

*/

// #####################################################
// Configuration
// #####################################################

namespace {
	require_once(dirname(__FILE__).'/conf.php');
	
	// Defaults:
	$CONF_DEFAULTS = array(
		'S_PAGETITLE'      => 'YATWiki',
		'S_BASEURL'        => '/',
		'S_DEFAULTPAGE'    => 'home',
		'S_TIMEZONE'       => 'UTC',
		'S_DATEFORMAT'     => 'Y M jS, g:i a',
		'DB_PATH'          => dirname(__FILE__).'/wiki.db',
		'DB_PATH_HTTP'     => S_BASEURL.'wiki.db',
		'ALLOW_DOWNLOAD'   => false,
		'NUM_RECENTCHANGES'=> 20,
		'GZ_COMPRESSION'   => 9
	);
	
	foreach($CONF_DEFAULTS as $cfd_k => $cfd_v) {
		if (! defined($cfd_k)) define($cfd_k, $cfd_v);
	}
	
	date_default_timezone_set(S_TIMEZONE);
	session_start();
	
	define('HTTP_ROOT', (!empty($_SERVER['HTTPS'])?'https':'http').'://'.$_SERVER['HTTP_HOST'].S_BASEURL);
	
	// Process bans
	
	if (! isset($BANS)) $BANS = array();
	foreach($BANS as $ip=>$binf) {
		if (preg_match($ip, $_SERVER['REMOTE_ADDR']) &&
		    (!$binf[2] || time() < $binf[2]) &&
			(!isset($_SESSION['evade'][$ip])||$_SESSION['evade'][$ip]!=$binf[1])
		) {
			if (isset($_GET['evade'])) {
				if ($_GET['evade']==$binf[1]) {
					$_SESSION['evade'][$ip] = $binf[1];
					die('ban evaded <a href="'.S_BASEURL.'">ok &raquo;</a>');
				} else die('incorrect');
			}
			die('<!DOCTYPE html>
<b>banned:</b> '.\h($binf[0]).'<br>
<b>expires:</b> '.($binf[2]? date(S_DATEFORMAT, $binf[2]): 'never').'<br>
<form method="get"><input type="text" name="evade" style="border:0;"></form>
			');
		}
	}
	
}

// #####################################################
// Database
// #####################################################

namespace DB {

	function dbh() {
		static $db = null;
		if (is_null($db)) {
			$db = new \PDO('sqlite:'.DB_PATH);
			$db->query('
				CREATE TABLE IF NOT EXISTS articles (
					id			INTEGER PRIMARY KEY,
					article		INTEGER,
					modified	INTEGER,
					body		TEXT,
					author		TEXT
				);
			');
			$db->query('
				CREATE TABLE IF NOT EXISTS titles (
					id INTEGER PRIMARY KEY,
					title TEXT
				);
			');
			$db->query('
				CREATE INDEX IF NOT EXISTS articles_modified_index
				ON articles (modified)
			');
		}
		return $db;
	}

	function exec($sql, array $args = array()) {
		$s = dbh()->prepare($sql);
		$s->execute($args);
		return $s->fetchAll(\PDO::FETCH_ASSOC);
	}
	
	function SearchTitle($title) {
		$results = exec(
			'SELECT id FROM titles WHERE title = ? LIMIT 1',
			array(strtolower(trim($title)))
		);
		return (count($results)) ? $results[0] : BlankArticle($title);
	}

	function listArticles() {
		return exec('SELECT title FROM titles ORDER BY title ASC');
	}
	
	function countEdits() {
		$ret = exec('SELECT COUNT(id) c FROM articles');
		return $ret[0]['c'];
	}
	
	function GetArticle($title) {
		$article = SearchTitle($title);
		if (! $article['id']) return BlankArticle($title);
		return GetLatestEditionOf($article['id']);
	}
	
	function GetLatestEditionOf($article_id) {
		$results = exec(
			'SELECT * FROM articles WHERE article=? ORDER BY modified DESC LIMIT 1',
			array($article_id)
		);
		if (count($results)) {
			$results[0]['body'] = @gzinflate($results[0]['body']);
			return $results[0];
		}
		return BlankArticle();
	}

	function GetRandomArticleTitle() {
		$results = exec('SELECT title FROM titles ORDER BY RANDOM() LIMIT 1');
		return $results[0]['title'];
	}

	function GetRecentArticles($count, $page=1) {
		$results = exec(
			'SELECT articles.*, titles.title title FROM articles '.
			'JOIN titles ON articles.article=titles.id '.
			'ORDER BY modified DESC LIMIT ?,?',
			array( max(0,$page-1)*$count, $count )
		);
		return (count($results)) ? $results : array();
	}

	function GetPreviousVersions($article_id) {
		$results = exec(
			'SELECT id, modified, author FROM articles '.
			'WHERE article=? ORDER BY modified DESC',
			array($article_id)
		);
		return (count($results)) ? $results : array();
	}
	
	function GetArticleVersionBefore($title_id, $article_id) {
		$results = exec(
			'SELECT id FROM articles WHERE article = ? AND id < ? ORDER BY id DESC LIMIT 1',
			array($title_id, $article_id)
		);
		return (count($results) == 1) ? $results[0]['id'] : null;
	}

	function GetRawArticle($history_id) {
		$results = exec(
			'SELECT articles.*, titles.title title FROM articles '.
			'JOIN titles ON articles.article=titles.id WHERE articles.id=?',
			array($history_id)
		);
		if (count($results)) {
			$results[0]['body'] = gzinflate($results[0]['body']);
			return $results[0];
		}
		return BlankArticle();
	}

	function BlankArticle($title = '') {
		return array(
			'id'=>0,
			'modified'=>time(),
			'author'=>'unknown',
			'title'=>strtolower(trim($title)),
			'body'=>''
		);
	}

	function SaveArticle($title, $body) {
		if (! strlen($title)) return;
		$article = SearchTitle($title);
		
		if (intval($article['id']) == 0) {
			exec('INSERT INTO titles (title) VALUES (?)', array(strtolower(trim($title))));
			$article['id'] = dbh()->lastInsertId();
		}

		exec(
			'INSERT INTO articles (article, modified, body, author) '.
			'VALUES (?, ?, ?, ?)',
			array(
				$article['id'], time(),
				gzdeflate($body, GZ_COMPRESSION),
				$_SERVER['REMOTE_ADDR'].'-'.
				substr(md5($_SERVER['HTTP_USER_AGENT']), 0, 6)
		));
	}

}

// #####################################################
// Controller + Views
// #####################################################

namespace {

	// Determine mode
	// ##############
	
	$GLOBALS['YTW_setTitle'] = '';
	$content = html_route(array(
		
		// Viewing page
		'~^/view/(.*)/$~' => function($title) {
			$title = urldecode($title);
			$GLOBALS['YTW_setTitle'] = $title;
			$article = DB\GetArticle($title);
			if ($article['id'] == 0) {
				if ($title == S_DEFAULTPAGE) redirect(S_BASEURL.'modify/'.uenc(S_DEFAULTPAGE));
				no_such_article($title);
			} else return displayfmt($article['body']);
		},
		'~^/view/(.*)$~' => function($title) {
			redirect(S_BASEURL.'view/'.$title.'/'); // no uenc
		},
		
		// Download source
		'~^/get-src/$~' => function() {
			if (!ALLOW_DOWNLOAD) home('Feature disabled.');
			
			header('Content-Type: application/octet-stream');
			header('Content-Disposition: attachment;filename="'.
			       'YATWiki-'.date('YmdHis', filemtime(__FILE__)).'.php.gz"'
			);
			die(gzencode(file_get_contents(__FILE__), GZ_COMPRESSION));
		},
		
		// Random page
		'~^/random/$~' => function() {
			redirect(S_BASEURL.'view/'.uenc(DB\GetRandomArticleTitle()).'/');
		},
		
		// Show differences
		'~^/diff/(\d+)/(\d+)/?$~' => function($f, $t) {
			$left  = DB\GetRawArticle(intval(min($f, $t)));
			$right = DB\GetRawArticle(intval(max($f, $t)));
			
			if ($left['title'] != $right['title']) {
				home('Invalid attempt to compare unrelated pages.');
			}
			
			return '<h2>Comparing "'.h($left['title']).'" versions '.
				'<a href="'.S_BASEURL.'archive/'.$left['id'].'">r'.$left['id'].'</a> - '.
				'<a href="'.S_BASEURL.'archive/'.$right['id'].'">r'.$right['id'].'</a>'.
				'</h2><pre>'.htmlDiff(h($left['body']),h($right['body'])).'</pre>'
			;
		},
		'~^/diff/.*$~' => function() {
			if (isset($_GET['t']) && isset($_GET['f'])) {
				redirect(S_BASEURL.'diff/'.intval($_GET['f']).'/'.intval($_GET['t']).'/');
			}
			home('Invalid request.');
		},
		
		// Recent changes
		'~^/recent/$~' => function() {
			redirect(S_BASEURL.'recent/1');
		},
		'~^/recent/(\d+)$~' => function($page) {
			$maxpage = ceil(DB\countEdits() / NUM_RECENTCHANGES);
			if ($page < 1) redirect(S_BASEURL.'recent/1');
			if ($page > $maxpage) redirect(S_BASEURL.'recent/'.$maxpage);
			$articles = DB\GetRecentArticles(NUM_RECENTCHANGES, $page);
			
			$content  = '<h2>Recent Changes</h2><br><em>Showing up to '.
				NUM_RECENTCHANGES.' recent changes.</em><br><br><table>';
				
			foreach($articles as $article) $content .= html_row(
				$article['title'], S_BASEURL.'view/'.uenc($article['title']),
				'[<a href="'.S_BASEURL.'archive/'.$article['id'].'">a</a>]',
				date(S_DATEFORMAT, $article['modified']).' by '.$article['author']
			);
			
			$content .= '<tr><td>';
			if ($page > 1) $content .= '<a href="'.S_BASEURL.'recent/'.($page-1).'">&#xab; Newer</a>';
			$content .= '</td><td></td><td style="text-align:right;">';
			if ($page < $maxpage) $content .= '<a href="'.S_BASEURL.'recent/'.($page+1).'">Older &raquo;</a>';
			$content .= '</td></tr></table>';
			return $content;
		},	
		
		// Article index
		'~^/index/$~' => function() {
			$articles = DB\ListArticles();
			$ret = array();
			$ret[] = 
				'<h2>Article Index</h2><br><em>There are '.
				(\DB\countEdits()).' edits to '.
				(count($articles)).' pages.</em><br><br><ul>'
			;
			foreach ($articles as $article) $ret[] =
				'<li><a href="'.S_BASEURL.'view/'.uenc($article['title']).'/">'.
				h($article['title']).'</a></li>'
			;
			$ret[] = '</ul>';
			return implode("\n", $ret);
		},
		
		// History
		'~^/history/(.*)/$~' => function($title) {
			$title = urldecode($title);
			$GLOBALS['YTW_setTitle'] = $title;
			
			$article = DB\SearchTitle($title);
			if ($article['id'] == 0) {
				no_such_article($title);
			} else {
				$archives = DB\GetPreviousVersions($article['id']);
				$content = '
					<h2>Page History</h2><br>
					<em>There have been '.count($archives).' edits to the page
							&quot;<a href="'.S_BASEURL.'view/'.uenc($title).'/">'.
							h($title).'</a>&quot;.
					</em><br><br>
					<form method="get" action="'.S_BASEURL.'diff/">
					<table>
				';
				$content .= ($COMPARE_ROW = '<tr><td colspan="2"></td><td><input type="submit" value="Compare Selected &raquo;"></td></tr>');
				foreach($archives as $arc) {
					$content .= html_row(
						date(S_DATEFORMAT, $arc['modified']),
						S_BASEURL.'archive/'.$arc['id'],
						$arc['author'],
						'<input type="radio" name="t" value="'.$arc['id'].'">'.
						'&nbsp;'.
						'<input type="radio" name="f" value="'.$arc['id'].'">'
					);
				}
				$content .= $COMPARE_ROW;
				$content .= '</table></form>';
			}
			return $content;
		},
		
		// Archived entry
		'~^/archive/(\d+)$~' => function($rev) {
			$article = DB\GetRawArticle($rev);
			if ($article['id'] == 0) {
				home('No such item in archive.');
			} else {
				$GLOBALS['YTW_setTitle'] = $article['title'];
				return '
					<div class="info">
					You are viewing a specific revision of this page,
					last modified '.date(S_DATEFORMAT, $article['modified']).'.
					Click <a href="'.S_BASEURL.'view/'.uenc($article['title']).'/">here</a>
					to see the latest revision.</div>
					'.
					displayfmt($article['body'])
				;
			}
		},
		
		// Raw text view
		'~^/raw/(\d+)$~' => function($rev) {
			$article = DB\GetRawArticle($rev);
			if (! $article['id']) { header('HTTP/1.0 404'); die(); }
			header('Content-Type: text/plain; charset=UTF-8');
			die($article['body']);
		},
		
		// RSS: changes
		'~^/rss/changes$~' => function() {
			$articles = DB\GetRecentArticles(5, 0);
			header('Content-Type: text/plain');
			echo '<?xml version="1.0" encoding="UTF-8" ?> 
<rss version="2.0" xmlns:atom="http://www.w3.org/2005/Atom"> 
	<channel>
		<title>'.h(S_PAGETITLE).'</title>
		<link>'.HTTP_ROOT.'</link>
		<description>'.h(S_PAGETITLE).'</description>
		<language>en-GB</language>
		<atom:link href="'.HTTP_ROOT.'rss/changes" rel="self" type="application/rss+xml" />';
			foreach($articles as $article) {
				// Get previous revision
				$previous = DB\GetArticleVersionBefore($article['article'], $article['id']);
				$post = '';
				if (! is_null($previous)) {
					$post = '| <a href="'.h(HTTP_ROOT.'diff/'.$article['id'].'/'.$previous).'">diff to previous r'.$previous.'</a>';
				}
				
				// RSS channel item
				echo '
		<item>
			<title>'.h($article['title']).' (r'.$article['id'].')</title>
			<link>'.HTTP_ROOT.'archive/'.$article['id'].'</link>
			<guid>'.HTTP_ROOT.'archive/'.$article['id'].'</guid>
			<author>nobody@example.com ('.h(S_PAGETITLE.' '.$article['author']).')</author>
			<pubDate>'.h(date(DATE_RSS, $article['modified'])).'</pubDate>
			<description>'.h(' 
				<a href="'.h(HTTP_ROOT.'view/'.uenc($article['title'])).'">latest version</a>
				| <a href="'.h(HTTP_ROOT.'archive/'.$article['id']).'">revision '.$article['id'].'</a>
				'.$post.'
			').'</description>
		</item>';
			}
			echo '
	</channel>
</rss>';
			die();
		},
		
		// BBCode help
		'~^/formatting/~' => function() {
			return '
<h2>Formatting help</h2><br><br>
<ul>
	<li>[h]header[/h]</li>
	<li>[b]bold[/b]</li>
	<li>[u]underline[/u]</li>
	<li>[i]italic[/i]</li>
	<li>[s]strikethrough[/s]</li>
	<li>[spoiler]spoiler[/spoiler]</li>
	<li>[list] item [*] item [/list]</li>
	<li>[url=address]title[/url]</li>
	<li>[article=page name]title[/article] or [rev=id]title[/rev]</li>
	<li>[img]image-url[/img]</li>
	<li>[imgur]asdf.jpg[/imgur]</li>
	<li>[code]fixed width[/code]</li>
	<li>[section=header]content[/section]</li>
	<li>[html]raw html[/html]</li>
</ul>
			';
		},
		
		// Edit/create page
		'~^/modify/(.*)/~' => function($title) {
			$title = urldecode($title);
			$article = DB\GetArticle($title);
			if (strlen($title)) {
				$GLOBALS['YTW_setTitle'] = $title;
				$cheader = '
Editing article &quot;<a href="'.S_BASEURL.'view/'.uenc($title).'">'.h($title).'</a>&quot;
				';
			} else $cheader = "Creating new article";
			
			return '
<h2>'.$cheader.'</h2><br>

<form method="POST" action="'.S_BASEURL.'save/" class="editor" accept-charset="UTF-8">
	<div class="frm">
		<label>
			Save as:
			<input type="hidden" name="baserev" value="'.$article['id'].'" />
			<input type="text" name="pname" value="'.h($title).'"/>
		</label>
		<input type="submit" value="Save &raquo;">
		| <a href="'.S_BASEURL.'formatting/" target="_blank">formatting&nbsp;help</a>
	</div>
	<div id="contentctr"><textarea name="content">'.h($article['body']).'</textarea></div>
</form>
			';
		},
		
		'~/save/$~' => function() { // POST
			foreach(array('pname', 'content', 'baserev') as $arg) {
				if (!isset($_POST[$arg])) home('Invalid request.');
			}
			
			if (!strlen($pname = trim($_POST['pname']))) {
				home('Missing page title.');
			}
			
			$prev = DB\GetArticle($pname);
			if (intval($prev['id']) !== intval($_POST['baserev'])) {
				home("Warning: Your changes were not based on the most recent ".
				     "version of the page (r".$prev['id']." &ne; r".
				     intval($_POST['baserev'])."). No changes were saved.",
					 S_BASEURL.'modify/'.uenc($pname)
				);
			}

			DB\SaveArticle($pname, $_POST['content']);
			redirect(S_BASEURL.'view/'.uenc($pname).'/');
		},
		
		'~^/$~' => function() {
			redirect();
		}
	), array(
		'default' => function() { home('Invalid request.'); }
	));
		

// #####################################################
// Helper functions
// #####################################################

	function h($s) { return htmlspecialchars($s); }
	function home($sz, $dest='') { $_SESSION['info'] = $sz; redirect($dest); }
	function uenc($data) { return str_replace('%2F', '/', urlencode($data)); }
	function xesc($sz) {
		return implode('', array_map(function($c) {
			return (ctype_alnum($c) || is_int(strpos("[](){}:;,.=*+-_ ",$c))) ? $c : sprintf('\x%02x', ord($c));
		}, str_split($sz)));
	}
	
	function no_such_article($title) {
		home('No such article exists. '.
		     '<a href="'.S_BASEURL.'modify/'.uenc($title).'">Click here</a> to create it.'
		);
	}
	
	function html_row($label, $url) {
		$args = func_get_args();
		$ret = '<tr><td><a href="'.$url.'">'.h($label).'</a></td>';
		if (count($args) > 2) for ($i = 2, $e = count($args); $i < $e; $i++) {
			$ret .= '<td>'.$args[$i].'</td>';
		}
		return $ret."</td>\n";
	}

	function redirect($target='') {
		if ($target=='') $target = S_BASEURL.'view/'.uenc(S_DEFAULTPAGE).'/';
		header('Location: '.$target);
		die();
	}
	
	function html_route($routes, $options=array()) {
		$options = array_merge(array(
			'request' => $_SERVER['REQUEST_URI'],
			'default' => function(){}
		), $options);		
		foreach($routes as $route => $handler) {
			$matches = array();
			if (preg_match($route, $options['request'], $matches)) {
				return call_user_func_array($handler, array_slice($matches, 1));
			}
		}		
		return call_user_func($options['default']);
	}
		
	function pregwipe($data, $s_to_r) {
		foreach($s_to_r as $s=>$r) {
			while(preg_match($s, $data)) {
				if (is_callable($r)) {$data = preg_replace_callback($s, $r, $data);
				} else {$data = preg_replace($s, $r, $data);}
			}
		}
		return $data;
	}

	// An embarassing cascade of half-working hacks follows.
	
	function bbtables($string) {
		$lines  = explode("\n", $string);
		$lastct = 0;
		$ret    = array();
		$tbl    = array();
		$buffer = array();
		foreach ($lines as $line) {
			if (strlen($line) && $line[0] == '|') {
				if (count($buffer)) {
					$ret[] = bbcode(h(implode("\n",$buffer))); // SYNCHRONISE
					$buffer = array();
				}
				
				$cols = explode('|', $line);
				if ($lastct != count($cols)) {
					if ($lastct > 0) { // Close previous table
						$tbl[] = '</table>';
						$ret[] = implode('', $tbl);
						$tbl = array();
						$lastct = 0;
					}
					// Start new table
					$lastct = count($cols);
					$tbl[] = '<table class="ti">';
				}
				$cols = array_map(function($c) {
					return bbcode(h(str_replace(';',"\n",$c))); // SYNCHRONISE
				}, $cols);
				// Add these columns
				$tbl[] = '<tr><td>'.implode('</td><td>',array_slice($cols, 1)).
					'</td></tr>'
				;
			} else {
				if ($lastct > 0) {
					// Close previous table
					$tbl[] = '</table>';
					$ret[] = implode('', $tbl);
					$tbl = array();
					$lastct = 0;
				}
				$buffer[] = $line;
			}
		}
		if (count($buffer)) {
			$ret[] = bbcode(h(implode("\n",$buffer))); // SYNCHRONISE
			$buffer = array();
		}
		if ($lastct > 0) {
			$tbl[] = '</table>';
			$ret[] = implode('', $tbl);
			$lastct = 0;
		}
		return implode("\n", $ret);
	}
	
	function bbcode($string) {
		return pregwipe($string, array(
			'@\[(?i)h\](.*?)\[/(?i)h\]@si'      => '<h2>\\1</h2>',
			'@\[(?i)b\](.*?)\[/(?i)b\]@si'      => '<b>\\1</b>',
			'@\[(?i)i\](.*?)\[/(?i)i\]@si'      => '<i>\\1</i>',
			'@\[(?i)u\](.*?)\[/(?i)u\]@si'      => '<u>\\1</u>',
			'@\[(?i)s\](.*?)\[/(?i)s\]@si'      => '<span class="s">\\1</span>',
			'@\[(?i)spoiler\](.*?)\[/(?i)spoiler\]@si' => '<span class="spoiler">\\1</span>',
			'@\[(?i)img\](.*?)\[/(?i)img\]@si'  => '<img alt="" src="\\1">',
			'@\[(?i)list\](.*?)\[/(?i)list\]@si'=> '<ul><li>\\1</li></ul>',
			'@\[\*\]@si'                        => '</li><li>',
			
			'@\[(?i)url=(.*?)\](.*?)\[/(?i)url\]@si' => '<a href="\\1">\\2</a>',
			
			'@\[(?i)article=(.*?)\](.*?)\[/(?i)article\]@si' => function($m) {
				return '<a href="'.S_BASEURL.'view/'.uenc($m[1]).'/">'.$m[2].'</a>';
			},			
			'@\[(?i)rev=(.*?)\](.*?)\[/(?i)rev\]@si' => function($m) {
				return '<a href="'.S_BASEURL.'archive/'.uenc($m[1]).'">'.$m[2].'</a>';
			},
			
			'@\[(?i)imgur\](.*?)\.(...)\[/(?i)imgur\]@si' =>
				'<a href="http://i.imgur.com/\\1.\\2">'.
				'<img class="imgur" alt="" src="http://i.imgur.com/\\1s.\\2" >'.
				'</a>'
			,
			
			'@\[(?i)section=(.*?)](.*?)\[/(?i)section\]@si' => function($m) {
				return '<div class="section"><a class="sectionheader" href="javascript:;" onclick="ts(this);">'.$m[1].'</a><span style="display:none;">'.trim($m[2]).'</span></div>';
			},

		));
	}
	
	function bbformat($str) {
		return preg_replace_callback(
			'@\[(?i)code\](.*?)\[/(?i)code\]@si',
			function($m) {
				$GLOBALS['codePresent'] = true;
				return '<pre><code>'.str_replace("<br>\r\n","\r\n",$m[1]).'</code></pre>';
			},
			str_replace(
				array("</ul><br>\r\n<br>"),
				array('</ul><br>'),
				nl2br(bbtables($str), false) //bbtables(bbcode(h($str)))
			)
		);
	}
	
	function displayfmt($string) {
		$hpos = 0; $ret = '';
		while (($spos = strpos($string, "[html]", $hpos)) !== false) {
			$ret .= bbformat(substr($string, $hpos, $spos-$hpos));
			$spos += 6;
			
			$epos = strpos($string, "[/html]", $spos);
			$ret .= '<div class="html"><a href="javascript:;" onclick="els(this, \''.xesc(substr($string, $spos, $epos-$spos)).'\');">&#x26a0; run dynamic content</a></div>';
			$hpos = $epos+7;
		}
		return $ret.bbformat(substr($string, $hpos));
	}

// #####################################################
// Third-party code
// #####################################################
	
	/*
	Paul's Simple Diff Algorithm v 0.1
	(C) Paul Butler 2007 <http://www.paulbutler.org/>
	May be used and distributed under the zlib/libpng license.
	https://github.com/paulgb/simplediff/
	*/

	function diff($old, $new){
		$maxl = 0;
		foreach($old as $oindex => $ovalue){
			$nkeys = array_keys($new, $ovalue);
			foreach($nkeys as $nindex){
				$matrix[$oindex][$nindex] =
					isset($matrix[$oindex - 1][$nindex - 1]) ?
					$matrix[$oindex - 1][$nindex - 1] + 1 : 1
				;
				if($matrix[$oindex][$nindex] > $maxl){
					$maxl = $matrix[$oindex][$nindex];
					$omax = $oindex + 1 - $maxl;
					$nmax = $nindex + 1 - $maxl;
				}
			}	
		}
		if($maxl == 0) return array(array('d'=>$old, 'i'=>$new));
		return array_merge(
			diff(array_slice($old, 0, $omax), array_slice($new, 0, $nmax)),
			array_slice($new, $nmax, $maxl),
			diff(array_slice($old,$omax+$maxl), array_slice($new,$nmax+$maxl)));
	}

	function htmlDiff($old, $new){
		$ret = '';
		$diff = diff(explode("\n", $old), explode("\n", $new));
		foreach($diff as $k) if(is_array($k)) {
			$ret .= (!empty($k['d'])?"<del>".implode("\n",$k['d'])."</del>":'').
			        (!empty($k['i'])?"<ins>".implode("\n",$k['i'])."</ins>":'');
		} else {
			$ret .= $k . "\n";
		}
		return $ret;
	}
	
	
// #####################################################
// Main Template
// #####################################################

header('Content-Type: text/html;charset=UTF-8');
header('X-Served-By: '.S_PAGETITLE);
?>
<!DOCTYPE html>
<html>
	<head>
		<title><?=h(S_PAGETITLE)?><?=(strlen($GLOBALS['YTW_setTitle'])?' - '.h($GLOBALS['YTW_setTitle']):'')?></title>		
		<meta charset="utf-8">
		<meta name="viewport" content="width=device-width">
		<meta http-equiv="X-UA-Compatible" content="IE=edge" />
		<link rel="alternate" type="application/rss+xml" href="<?=S_BASEURL?>rss/changes" title="<?=h(S_PAGETITLE)?> - Recent Changes">
		<link rel="stylesheet" type="text/css" href="<?=S_BASEURL?>wiki.css?_=<?=filemtime('wiki.css')?>">
<?php if (array_key_exists('codePresent', $GLOBALS)) { ?> 
		<script type="text/javascript" src="<?=S_BASEURL?>highlight.js"></script>
		<script type="text/javascript">hljs.tabReplace = '    ';hljs.initHighlightingOnLoad();</script>
<?php } ?> 
		<script type="text/javascript">
var a=document;
function tid(id){t(a.getElementById(id));};function ts(e){t(e.nextSibling);}
function t(e){e.style.display=(e.style.display=='none')?'block':'none';}
function els(e,s){ <?php // no js exec in innerHTML ?> 
 var p=e.parentNode;p.className="";p.innerHTML=s;n="script";
 var z=p.childNodes,m="text/javascript",l=function(s){return s.toLowerCase();};
 for (var i=0,e=0;e=z[i];i++) {
  if (e.nodeName&&(l(e.nodeName)===n)&&(!e.type||l(e.type)===m)){
   var t=p.removeChild(e),d=a.getElementsByTagName("head")[0],
       se=a.createElement(n);se.type=m;
   se.appendChild(a.createTextNode((t.text||t.textContent||t.innerHTML||"")));
   d.insertBefore(se,d.firstChild ); }}}
		</script>
	</head>
	<body>
		<div class="header">
			<a href="<?=S_BASEURL?>view/<?=uenc(S_DEFAULTPAGE)?>/" title="Home"><div class="sprite hm"></div></a>
			<a href="javascript:;" onclick="tid('spm');tid('tr1');tid('tr2');" title="Menu"><div class="sprite sp"></div></a>
			<a href="<?=S_BASEURL?>modify/untitled-<?=time()?>/" title="New Page"><div class="sprite nw"></div></a>
<?php if (strlen($GLOBALS['YTW_setTitle'])) { ?>
			<div class="sep"></div>
			<a href="<?=S_BASEURL?>history/<?=uenc($GLOBALS['YTW_setTitle'])?>/" title="Page History"><div class="sprite hs"></div></a>
			<a href="<?=S_BASEURL?>modify/<?=uenc($GLOBALS['YTW_setTitle'])?>/" title="Modify Page"><div class="sprite ed"></div></a>
<?php } ?> 
		</div>
		<div id="tr1" style="display:none;"></div>
		<div id="tr2" style="display:none;"></div>
		<div class="ddmenu" id="spm" style="display:none;">
			<a href="<?=S_BASEURL?>recent/"><div class="sprite no"></div> Recent Changes</a>
			<a href="<?=S_BASEURL?>random/"><div class="sprite rn"></div> Random Page</a>
			<a href="<?=S_BASEURL?>index/"><div class="sprite no"></div> Article Index</a>
<?php if (ALLOW_DOWNLOAD) { ?>
			<a href="<?=DB_PATH_HTTP?>"><div class="sprite no"></div> Download DB backup</a>
			<a href="<?=S_BASEURL?>get-src/"><div class="sprite no"></div> Download Wiki Source</a>
<?php } ?>
		</div>
		<div class="content">
<?php if (isset($_SESSION['info'])) { ?>
			<div class="info"><?=$_SESSION['info']?></div>
<?php unset($_SESSION['info']);
	  }
?>
<?=$content?> 
		</div>
	</body>
</html>
<?php
} // namespace
